-- This game is partially based on the instructions found on:
-- http://elm-lang.org/blog/pong


module Main (..) where

import Graphics.Collage exposing (..)
import Graphics.Element exposing (..)
import Time
import Keyboard
import Mouse
import Text
import Window
import Random exposing (..)
import Color exposing (..)
import Render exposing (..)
import Data exposing (..)
import Physics exposing (..)
import Powerup exposing (..)
import AnimationFrame


-------------------------------------------------------------------------------
-----------------------------------------Main----------------------------------
-------------------------------------------------------------------------------


main : Signal Element
main =
  Signal.map2 render gameState Window.dimensions



-------------------------------------------------------------------------------
-----------------------------------------Timer---------------------------------
-------------------------------------------------------------------------------


toggle : Signal Bool -> Signal Bool
toggle signal =
  Signal.foldp (\_ b -> not b) True (Signal.filter not False (Signal.dropRepeats signal))


type alias Timer =
  { time : Time.Time
  , input : Input
  }


type alias Input =
  { space : Bool
  , mouse : ( Int, Int )
  , mouseClick : Bool
  , window : ( Int, Int )
  , dir : ( Int, Int )
  }


timer : Signal Timer
timer =
  Keyboard.arrows
    |> Signal.map (\{ x, y } -> ( x, y ))
    |> Signal.map5 Input Keyboard.space Mouse.position Mouse.isDown Window.dimensions
    |> Signal.map2 Timer clock
    |> Signal.sampleOn clock


clock : Signal Float
clock =
  Signal.map Time.inSeconds (AnimationFrame.frameWhen (toggle Keyboard.shift))



-------------------------------------------------------------------------------
----------------------------------------Update---------------------------------
-------------------------------------------------------------------------------


gameState : Signal GameState
gameState =
  Signal.foldp updateGame defaultGame timer


updateGame : Timer -> GameState -> GameState
updateGame timer gameState =
  case gameState.state |> Debug.watch "State" of
    NewGame ->
      defaultGame

    --if timer.input.mouseClick || timer.input.space then defaultGame else gameState
    NewLevel ->
      updateGameNewLevel gameState

    Play ->
      updateGamePlay timer gameState

    NewBall ->
      updateGameNewBall timer gameState

    Highscore ->
      updateHighscore timer gameState


updateHighscore : Timer -> GameState -> GameState
updateHighscore timer ({ gameOver, state, player, ballsList, bricksList, levels, levelNum, seed } as gameState) =
  let
    newState =
      if timer.input.mouseClick || timer.input.space then
        NewGame
      else
        Highscore
  in
    { gameState
      | state = NewBall
      , ballsList = []
      , levels = []
      , fallingPowerups = []
      , activePowerups = []
    }


updateGameNewLevel : GameState -> GameState
updateGameNewLevel ({ gameOver, state, player, ballsList, bricksList, levels, levelNum, seed } as gameState) =
  let
    newState =
      NewBall

    ( newBall, seed' ) =
      getRandomBall seed

    ( newBricksList, newGameOver ) =
      case List.head levels of
        Just bricks ->
          ( uncurry translateMatrixToBricklist bricks, gameOver )

        Nothing ->
          ( bricksList, True )

    newLevels =
      case List.tail levels of
        Just bricksList ->
          bricksList

        Nothing ->
          []
  in
    { gameState
      | state = NewBall
      , gameOver = newGameOver
      , player = player
      , ballsList = [ newBall ]
      , bricksList = newBricksList
      , levels = newLevels
      , levelNum = levelNum + 1
      , seed = seed'
      , fallingPowerups = []
      , activePowerups = []
    }


updateGameNewBall : Timer -> GameState -> GameState
updateGameNewBall ({ input, time } as timer) ({ gameOver, state, player, ballsList, bricksList, seed } as gameState) =
  let
    newGameOver =
      player.lives <= 0 || gameOver

    newState =
      if newGameOver then
        Highscore
      else
      --NewGame else
      if
        input.mouseClick || input.space
      then
        Play
      else
        NewBall

    newPlayer =
      updatePlayer player input time player.warp

    ( newBall, newSeed ) =
      getRandomBall seed

    newBricksList =
      bricksList
  in
    if gameOver && (input.mouseClick || input.space) then
      defaultGame
    else
      { gameState
        | state = newState
        , gameOver = newGameOver
        , player = newPlayer
        , ballsList = [ newBall ]
        , bricksList = newBricksList
        , seed = newSeed
      }


updateGamePlay : Timer -> GameState -> GameState
updateGamePlay ({ input, time } as timer) ({ gameOver, state, player, ballsList, bricksList, fallingPowerups, activePowerups, seed } as gameState) =
  let
    ( powAlt, seed' ) =
      getPowerupAlterations activePowerups defaultPowerupAlterations seed

    newState =
      if List.isEmpty bricksList then
        NewLevel
      else if List.isEmpty ballsList then
        NewBall
      else
        Play

    newPlayer =
      updatePlayer player input time powAlt.warp
        |> \tempPlayer ->
            if List.isEmpty ballsList then
              { tempPlayer | lives = player.lives - 1 }
            else
              tempPlayer
                |> \x ->
                    { x | shape = powAlt.playerShape }
                      |> alterPlayerSize powAlt.sizeMult
                      |> \x -> { x | lives = x.lives + powAlt.extraLife }

    ( newBallsList, tempBricksList' ) =
      updateBallsList ballsList newPlayer time powAlt
        |> \tempBallsList -> updateBallsBricksList tempBallsList bricksList

    tempBricksList =
      if powAlt.shoot then
        shootBrick tempBricksList' input
      else
        tempBricksList'

    ( newBricksList, scoreDelta, tempFallingPowerups, seed'' ) =
      deleteFailedBricks ( tempBricksList, 0, fallingPowerups, seed' )

    newScore =
      newPlayer.score + scoreDelta * powAlt.scoreMult - time * 3

    ( newFallingPowerups, tempActivePowerups ) =
      updatePowerups tempFallingPowerups activePowerups player time

    newActivePowerups =
      if (List.length bricksList <= 8) && List.isEmpty (List.filter (\x -> x.power == Shoot) tempActivePowerups) then
        (hammertime :: tempActivePowerups)
      else
        tempActivePowerups
  in
    { gameState
      | state = newState
      , player = { newPlayer | score = newScore }
      , ballsList = (List.append powAlt.newBalls newBallsList) |> slowDownBallsList powAlt.ballVeloMult
      , bricksList = newBricksList
      , fallingPowerups = newFallingPowerups
      , activePowerups = newActivePowerups
      , seed = seed''
    }



----------------------------------update Player -----------------------------


updatePlayer : Player -> Input -> Time.Time -> Bool -> Player
updatePlayer player input time warp =
  { player
    | coord =
        ( toFloat (fst input.mouse - (fst input.window) // 2), snd player.coord )
        -- toFloat( -( snd mouse - (snd window)//2)))  --|> Debug.watch "coord"
        -- velo = ((fst (input.dir |> Debug.watch "dir") |> toFloat)*playerVelo, (snd input.dir |> toFloat)*playerVelo)
  }
    |> boundCoord borders
    |> \x -> updateCoord x time |> \x -> { x | warp = warp }


alterPlayerSize : Float -> Player -> Player
alterPlayerSize mult ({ shape } as player) =
  case shape of
    Rectangle ( x, y ) ->
      { player | shape = Rectangle ( x * mult, y ) }

    Circle radius ->
      { player | shape = Circle (radius * mult) }



-----------------------------------update BallsList------------------------


updateBallsList : List Ball -> Player -> Time.Time -> PowerupAlterations -> List Ball
updateBallsList ballsList player time powAlt =
  List.map (updateBall time player powAlt) ballsList
    |> deleteFailedBalls


deleteFailedBalls : List Ball -> List Ball
deleteFailedBalls ballsList =
  case ballsList of
    x :: xs ->
      if snd x.coord < -halfHeight - borderWidth then
        deleteFailedBalls xs
      else
        x :: (deleteFailedBalls xs)

    [] ->
      []


updateBall : Time.Time -> Player -> PowerupAlterations -> Ball -> Ball
updateBall time player powAlt ball =
  updateCoord ball time
    |> updateCollWalls powAlt
    |> \newBall ->
        updateCollPlayer newBall player
          |> \nBall ->
              if powAlt.godmode then
                { nBall | godmode = powAlt.godmode, colour = red }
              else
                { nBall | godmode = powAlt.godmode, colour = black }


slowDownBallsList : Float -> List Ball -> List Ball
slowDownBallsList slowdown ballsList =
  case ballsList of
    ball :: restBalls ->
      (slowdownBall ball slowdown) :: (slowDownBallsList slowdown restBalls)

    [] ->
      []


slowdownBall : Ball -> Float -> Ball
slowdownBall ball slowdown =
  let
    xv =
      fst ball.velo

    yv =
      snd ball.velo

    velo =
      xv ^ 2 + yv ^ 2 |> sqrt

    factor =
      (ballVelo * slowdown) / velo

    newX =
      xv * factor

    newY =
      yv * factor
  in
    { ball | velo = ( newX, newY ) }



---------------------------------update Bricks-------------------------------


updateBallsBricksList : List Ball -> List Brick -> ( List Ball, List Brick )
updateBallsBricksList ballsList bricksList =
  case ballsList of
    ball :: restBalls ->
      updateBricksList ball bricksList 9999
        -- : (Ball, List Brick)
        |>
          \( newBall, newBricksList ) ->
            let
              ( newRestBalls, newestBricksList ) =
                updateBallsBricksList restBalls newBricksList
            in
              ( newBall :: newRestBalls, newestBricksList )

    [] ->
      ( [], bricksList )


updateBricksList : Ball -> List Brick -> Float -> ( Ball, List Brick )



--distance ensures that the balls direction is altered only by the nearest brick


updateBricksList ball bricksList distance =
  case bricksList of
    brick :: restBricks ->
      updateCollBrick ball brick distance
        |> \( newBall, newBrick, newDistance ) ->
            let
              ( newestBall, newRestBricks ) =
                updateBricksList newBall restBricks newDistance
            in
              ( newestBall, newBrick :: newRestBricks )

    [] ->
      ( ball, bricksList )


deleteFailedBricks : ( List Brick, Float, List Powerup, Seed ) -> ( List Brick, Float, List Powerup, Seed )
deleteFailedBricks ( bricksList, score, fallingPowerup, seed ) =
  case bricksList of
    brick :: restBricks ->
      if brick.armor <= 0 then
        let
          ( powerupList, seed' ) =
            getRandomPowerup brick seed
        in
          deleteFailedBricks
            ( restBricks
            , score + (toFloat brick.points)
            , List.append powerupList fallingPowerup
            , seed'
            )
      else
        deleteFailedBricks ( restBricks, score, fallingPowerup, seed )
          |> \( x, y, z, v ) -> ( brick :: x, y, z, v )

    [] ->
      ( [], score, fallingPowerup, seed )


shootBrick : List Brick -> Input -> List Brick
shootBrick xs input =
  List.map
    (\brick ->
      if collideMouse input brick then
        { brick | armor = brick.armor - 1 }
      else
        brick
    )
    xs


collideMouse : Input -> Brick -> Bool
collideMouse input brick =
  case brick.shape of
    Circle r ->
      False

    Rectangle ( width, height ) ->
      let
        x =
          toFloat (fst input.mouse - (fst input.window) // 2)

        y =
          toFloat (-(snd input.mouse - (snd input.window) // 2))

        ( xb, yb ) =
          brick.coord
      in
        if x > xb - width / 2 && x < xb + width / 2 then
          if y > yb - height / 2 && y < yb + height / 2 then
            True
          else
            False
        else
          False
